# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

"""
Parsing for display widgets.

We use the `displaylang` library to build a DisplayLang processor that accepts
those SymPy callables for which we have so far defined `AllowedCallable`s.
The set is expected to grow over time in response to user demand.
"""

from displaylang.build import make_displaylang_processor
from sympy.core.function import UndefinedFunction
from sympy.core.relational import Relational

from pfsc_examp.config import config
from .excep import ParsingExcep

from .sympy_allowed import (
    sympy_basic_allowed_callables,
    generalized_builtin_allowed_callables,
    sympy_other_allowed_callables,
    sympy_available_constants,
)


# The "basic vars" are those that are available for all builds.
# Here we combine the constants we've made available, along with
# the "sympy basic" and "generalized builtin" allowed callables.
basic_vars = {**sympy_available_constants}
for c in sympy_basic_allowed_callables + generalized_builtin_allowed_callables:
    basic_vars[c.name] = c.callable

all_allowed_callables = (
    generalized_builtin_allowed_callables +
    sympy_basic_allowed_callables +
    sympy_other_allowed_callables
)

displaylang_processor = make_displaylang_processor(
    basic_vars, all_allowed_callables, add_builtins=True,
    allow_local_var_calls=False,
    abstract_function_classes=[UndefinedFunction],
    abstract_relation_classes=[Relational],
    max_len=config["MAX_DISPLAY_BUILD_LEN"],
    max_depth=config["MAX_DISPLAY_BUILD_DEPTH"],
)


###############################################################################
# Bounds on arithmetic operations
#
# For now none of this is doing anything.
# Keeping here for possible future use.

EXPONENT_CAP = 100
INT_DIGIT_CAP = 200


def pow_check(b, e):
    if isinstance(b, int) and isinstance(e, int):
        # There is a largest exponent that we will allow, regardless of the base:
        if abs(e) > EXPONENT_CAP:
            raise ParsingExcep(f'Exponent too large: {e}.')
        # For smaller exponents we allow larger bases, but we don't want the total
        # number of digits in the power to grow too large. We estimate this as the
        # number of decimal digits in the base, times the exponent itself.
        if len(str(b)) * abs(e) > INT_DIGIT_CAP:
            raise ParsingExcep(f'Power too large: {b}^{e}.')


def mult_check(a, b):
    if isinstance(a, int) and isinstance(b, int):
        if len(str(a)) + len(str(b)) > INT_DIGIT_CAP:
            raise ParsingExcep(f'Product too large: {a}*{b}.')



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from displaylang.exceptions import ControlledEvaluationException

from pfsc_examp.calculate import calculate
from pfsc_examp.excep import MalformedExampImport, MissingExport
from pfsc_examp.parse.display import displaylang_processor
from pfsc_examp.util import adapt


def parse_imports(s):
    """
    Parse an import string of the kind that appears as key in the `import`
    dictionary of a DispWidget.

    The format of this string is a comma-delimited list of imports, where each
    import at least gives the name of a variable as exported from the DispWidget
    in question, and may also give a local name we wish to use for it after
    importing, separated by keyword `as`.

    In other words, an import string matches this grammar:

        import_string := import ("," import)*
        import := CNAME ("as" CNAME)?
        %skip whitespace

    Thus, imports can be as simple as,

        import: {
            'f': some_disp
        }

    importing `f` as 'f' from `some_disp`, or as complex as,

        import: {
            'f as g, a, b, alpha as zeta': some_disp
        }

    importing `f` as 'g', `a` as 'a', `b` as 'b', and `alpha` as 'zeta'.

    Note that there is no risk of key collision in such an `import` dictionary,
    because it does not make sense to import more than one thing under the same
    local name.

    :param s: the import string to be parsed
    :return: list of ordered pairs (exported name, local name)
    """
    def err():
        msg = f'Malformed imports: {s}'
        raise MalformedExampImport(msg)
    imps = [p.strip() for p in s.split(',')]
    pairs = []
    for imp in imps:
        words = imp.split()
        n = len(words)
        if n == 1:
            w = words[0]
            pairs.append((w, w))
        elif n == 3:
            exp, prep, loc = words
            if prep != 'as':
                err()
            pairs.append((exp, loc))
        else:
            err()
    return pairs


def make_disp(parent, info):
    params = info.get('params', {})
    imports = info.get('imports', {})
    build_code = info['build']
    export_names = info.get('export', None)
    return ExampDisplay(parent, params, imports, build_code, export_names)


class ExampDisplay:

    def __init__(self, parent, params, imports, build_code, export_names=None):
        """
        Build an `ExampDisplay` instance based on a *built* display widget.
        In other words, the args passed here are not identical with those a
        user writes when defining a display in a pfsc module; they are instead
        as processed by a `pfsc.lang.widgets.DispWidget()` at build time.
        In particular, while users put all imports together in a single
        `imports` dictionary, in a built display widget these are separated
        into params, and imports from other displays.

        :param parent: DispWidget
            The widget that is constructing this display.
        :param params: dict
            Maps local param names we wish to use, to the absolute libpaths of
            ParamWidgets representing those parameters.
        :param imports: dict
            Specifies exports to import from DispWidgets.
            The values are absolute libpaths of DispWidgets. The keys are
            strings specifying vars to import, as defined in the
            `parse_imports()` function.
        :param build_code: str or list[str]
            Python code that builds (a) our HTML display, and (b) any values we
            want to export.
        :param export_names: None or list[str]
            Optional list of names of vars defined in the build_code. If given,
            then *only* these vars will be exported, i.e. made available for
            other displays to import. If ``None``, then *all* vars defined in
            the build_code will be exported.
        """
        self.parent = parent
        self.params = params
        self.imports = imports
        self.build_code = self.to_code_string(build_code)
        self.export_names = export_names

        self.param_values = {}
        self.import_values = {}

        self._exports = {}
        self._html = None

        self.last_attempted_raw_value = None

    def getUid(self):
        return self.parent.getUid()

    @staticmethod
    def to_code_string(code):
        """
        Convert what may be a string, or a list of strings, into a single code
        string.
        """
        return '\n'.join(code) if isinstance(code, list) else code

    def obtain_param_values(self):
        return {
            name: self.parent.get_generator(lp).value
            for name, lp in self.params.items()
        }

    def obtain_import_values(self):
        values = {}
        for imp_str, lp in self.imports.items():
            disp = self.parent.get_generator(lp)
            pairs = parse_imports(imp_str)
            for exp, loc in pairs:
                if exp not in disp.exports:
                    raise MissingExport(f'Display {lp} is missing export "{exp}".')
                values[loc] = disp.exports[exp]
        return values

    @property
    def exports(self):
        return self._exports

    @property
    def html(self):
        return self._html

    def write_html(self):
        """
        The reason for having this method (which duplicates the @property method
        `html`) is to have a uniform interface implemented by both `ExampDisplay`
        and `Parameter`. The reason `Parameter` needs this to be a true method
        instead of a property method is so that exceptions will not be silenced
        by Pyodide. See note in `Parameter.write_html()`.
        """
        return self._html

    def build(self, raw=None):
        raw = adapt(raw)
        if raw is not None:
            self.last_attempted_raw_value = raw = self.to_code_string(raw)

        self.param_values = self.obtain_param_values()
        self.import_values = self.obtain_import_values()
        existing_values = {**self.param_values, **self.import_values}

        html = None
        exports = {}
        if raw is None and self.last_attempted_raw_value is not None:
            # When attempting to build based on a previously attempted value,
            # we simply fail silently in case of a ControlledEvaluationException.
            # The user has not actively entered a value, so should be happy to
            # have the display switch back to anything that works.
            code = self.last_attempted_raw_value
            try:
                html, exports = calculate(
                    displaylang_processor.process,
                    code, existing_values
                )
            except ControlledEvaluationException:
                pass
        if html is None:
            # Here we are either attempting to build the value passed by the
            # user, or our default code. In either case, we do NOT catch
            # exceptions, because the user needs to know about any errors.
            code = self.build_code if raw is None else raw
            html, exports = calculate(
                displaylang_processor.process,
                code, existing_values
            )

        self._html = (
            f'<div class="display">\n{html}\n</div>\n'
        )
        if self.export_names is not None:
            exports = {k: v for k, v in exports.items() if k in self.export_names}
        self._exports = exports



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

"""
Controlled calculation for the server environment.
"""

from pfsc_util.imports import from_import

current_app = from_import('flask', 'current_app')

retry = from_import('tenacity', 'retry')
RetryError = from_import('tenacity', 'RetryError')
wait_exponential = from_import('tenacity', 'wait_exponential')
retry_if_result = from_import('tenacity', 'retry_if_result')
stop_after_delay = from_import('tenacity', 'stop_after_delay')

get_rqueue = from_import('pfsc.rq', 'get_rqueue')
MATH_CALC_QUEUE_NAME = from_import('pfsc.constants', 'MATH_CALC_QUEUE_NAME')
PfscExcep, PECode = from_import('pfsc.excep', ['PfscExcep', 'PECode'])


# Some calculations may return `None` as their legitimate result
# (e.g. `primitive_root(8)`), so we need a special object type to
# represent "no result":
class NoResult:
    pass


def calculate(f, *args, **kwargs):
    """
    Use this function to wait for a calculation to be carried out by an RQ worker,
    with a timeout, i.e. maximum execution time before the job is halted and marked
    as failed.

    Note that this function blocks: it won't return until the job is either
    completed or timed out. Thus, the purpose of using a worker here is not to make
    the job asynchronous, but to have the job carried out in a separate process,
    which can be timed out, and which can be containerized.
    """
    def give_up(due_to_timeout=False):
        msg = f'Math calc job failed for {f}'
        code = PECode.MATH_TIMEOUT_EXPIRED if due_to_timeout else PECode.MATH_CALCULATION_FAILED
        return PfscExcep(msg, code)

    overall_timeout = current_app.config.get("MATH_JOB_QUEUE_TIMEOUT", 180)

    @retry(
        retry=retry_if_result(lambda r: isinstance(r, NoResult)),
        wait=wait_exponential(multiplier=0.02, exp_base=1.4, max=1),
        stop=stop_after_delay(overall_timeout),
    )
    def check_job(job):
        job.get_status(refresh=True)
        if job.is_finished:
            return job.result
        # When jobs time out, RQ marks them as failed.
        # (See `rq.registry.StartedJobRegistry.cleanup()`.)
        # However, to be safe we also catch stopped jobs.
        elif job.is_failed or job.is_stopped:
            raise give_up()
        return NoResult()

    q = get_rqueue(MATH_CALC_QUEUE_NAME)
    calc_timeout = current_app.config.get("MATH_CALCULATION_TIMEOUT", 3)
    job = q.enqueue_call(f, args=args, kwargs=kwargs, timeout=calc_timeout)
    try:
        result = check_job(job)
    except RetryError:
        result = NoResult()
    if isinstance(result, NoResult):
        raise give_up(due_to_timeout=True)
    else:
        return result



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from sympy import Integer
from sympy.ntheory import is_primitive_root, primitive_root
import jinja2

from pfsc_examp.calculate import calculate
from pfsc_examp.canon_arithmeticus import all_prim_res_under_100, one_prim_res_under_1000
from pfsc_examp.parameters.base import Parameter
from pfsc_examp.excep import MalformedParamRawValue


class PrimRes_Param(Parameter):
    """
    Primitive residue in elementary number theory

    ptype: "PrimRes"

    Value type: Integer

    Default: int

    Args:
        Required:
            m: int-valued
                The modulus.
        Optional: None.
    """

    arg_spec = {
        "REQ": {
            'm': {
                'type': Integer,
            },
        },
    }

    def __init__(self, parent, tex=None,
                 default=None, descrip=None, params=None,
                 args=None, last_raw=None, **other):
        super().__init__(parent, tex, default, descrip, params, args, last_raw)
        self.m = None
        # Depending on how big the modulus m is, we might store a list of all
        # of its primitive residues, or we might just store one primitive residue;
        # we might not store either of these if m is too large.
        self.all_options = []
        self.one_option = None

    def prebuild(self):
        # Read args
        self.m = self.resolved_args['m']
        m = self.m.value
        # If the modulus m is < 100, we can supply the set of all possible values
        # for this primitive residue.
        # If 100 < m < 1000 we can offer one value, and give the user a text box
        # in which to set a different value.
        # Otherwise we offer no suggestion.
        if m < 100:
            self.all_options = all_prim_res_under_100.get(m, [])
        elif m < 1000:
            self.one_option = one_prim_res_under_1000.get(m, None)

    def auto_descrip(self, include_value=True, editable=True):
        return '%s a primitive residue mod $%s$' % (
            self.write_name_and_value(include_value, editable),
            self.m
        )

    def auto_build(self):
        if self.one_option:
            return self.one_option
        elif self.all_options:
            return self.all_options[0]
        else:
            return calculate(primitive_root, self.m.value)

    def build_from_raw(self, raw):
        n = Integer(self.check_int(raw))
        m = self.m.value
        def reject():
            msg = f'{n} is not a generator mod {m}.'
            raise MalformedParamRawValue(msg, self)
        # If m < 100 then we can check directly from the list of all generators.
        if m < 100:
            if n % m not in self.all_options:
                reject()
            return n
        else:
            if not calculate(is_primitive_root, n, m):
                reject()
            return n

    def write_chooser_widget(self):
        ready_options = []
        free_option = True
        if self.all_options:
            ready_options = self.all_options
            free_option = False
        elif self.one_option:
            ready_options = [self.one_option]
        context = {
            'descrip': self.auto_descrip(),
            'selected_res': self.value,
            'ready_options': ready_options,
            'free_option': free_option,
        }
        return prim_res_chooser_widget_template.render(context)

prim_res_chooser_widget_template = jinja2.Template("""
<div class="chooser radio_panel_chooser prim_res_chooser">
    <div class="heading">{{ descrip }}</div>
    <div class="error_display"></div>
    {# #}
    {% if ready_options %}
        {# Buttons for ready-made options #}
        <div class="radio_panel">
            {% for r in ready_options %}
                {% if r == selected_res %}
                    <div tabindex="0" value="{{ r }}" class="radio_panel_button rpb_selected">{{ r }}</div>
                {% else %}
                    <div tabindex="0" value="{{ r }}" class="radio_panel_button">{{ r }}</div>
                {% endif %}
            {% endfor %}
        </div>
    {% endif %}
    {# #}
    {% if free_option %}
        {# Text field for any other values #}
        {% if selected_res not in ready_options %}
            <input class="textfield" type="text" placeholder="Other" value="{{selected_res}}"/>
        {% else %}
            <input class="textfield" type="text" placeholder="Other"/>
        {% endif %}
    {% endif %}
</div>
""")



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from sympy.core.numbers import igcd, Integer
import jinja2

from pfsc_examp.calculate import calculate
from pfsc_examp.parameters.base import Parameter
from pfsc_examp.contexts import ContextNames
from pfsc_examp.excep import MalformedParamRawValue


class Integer_Param(Parameter):
    """
    Rational integer.

    ptype: "Integer"

    Value type: Integer

    Default: int

    Args:
      Required: None.
      Optional:
        coprimeTo: int-valued
            Must be coprime to this.
        dividing: int-valued
            Must divide this.
        gt: int-valued
            Must be greater than this.
        lt: int-valued
            Must be less than this.
    """
    
    arg_spec = {
        "OPT": {
            'coprimeTo': {
                'type': Integer,
            },
            'dividing': {
                'type': Integer,
            },
            'gt': {
                'type': Integer,
            },
            'lt': {
                'type': Integer,
            },
        }
    }

    def __init__(self, parent, tex=None,
                 default=None, descrip=None, params=None,
                 args=None, last_raw=None, **other):
        super().__init__(parent, tex, default, descrip, params, args, last_raw)

    def get_name_for_extra(self, predicate_name):
        a = self.resolved_args.get(predicate_name, None)
        if a is None:
            return None
        else:
            return str(a)

    def auto_build(self):
        # We're not writing a sat solver!
        # If you defined conditions, you should provide a default value that
        # satisfies them.
        return self.build_from_raw(0)

    def build_from_raw(self, raw):
        n = self.check_int(raw)

        # Note: when formatting error messages, we use the _name_ of the parameter
        # to which this one is supposed to relate in a given way; not the current
        # value. This is more informative for the user.

        D = self.get_arg_value('dividing')
        C = self.get_arg_value('coprimeTo')
        G = self.get_arg_value('gt')
        L = self.get_arg_value('lt')

        D_name = self.get_name_for_extra('dividing')
        C_name = self.get_name_for_extra('coprimeTo')
        G_name = self.get_name_for_extra('gt')
        L_name = self.get_name_for_extra('lt')

        if D is not None and D % n != 0:
            raise MalformedParamRawValue(f'Must divide ${D_name}$.', self)

        if C is not None and calculate(igcd, C, n) != 1:
            raise MalformedParamRawValue(f'Must be coprime to ${C_name}$.', self)

        if G is not None and not n > G:
            raise MalformedParamRawValue(f'Must be greater than ${G_name}$.', self)

        if L is not None and not n < L:
            raise MalformedParamRawValue(f'Must be less than ${L_name}$.', self)

        return Integer(n)

    def auto_descrip(self, include_value=True, editable=True):
        # Start by getting any optional, related parameters.
        dividing_str = self.get_name_for_extra('dividing')
        coprime_str = self.get_name_for_extra('coprimeTo')
        lb_str = self.get_name_for_extra('gt')
        ub_str = self.get_name_for_extra('lt')

        # Build the symbolic part
        symbolic_part = self.write_name_and_value(include_value=include_value, editable=editable)
        if not include_value:
            # If we do not want to include the value, then the symbolic
            # section can look like one of the following:
            #   $n$
            #   $n > l$
            #   $n < u$
            #   $l < n < u$
            if lb_str and not ub_str:
                symbolic_part += f' &gt; {lb_str}'
            elif ub_str and not lb_str:
                symbolic_part += f' &lt; {ub_str}'
            elif lb_str and ub_str:
                symbolic_part = f'{lb_str} $lt; {symbolic_part} &lt; {ub_str}'
            symbolic_part = f'${symbolic_part}$'

        # Set noun phrase according to context.
        noun_phrase = ('a rational integer'
                       if self.parent.context == ContextNames.AlgNT
                       else 'an integer')

        # Prepositional phrases:
        prep = []
        if include_value:
            # If we included the value, then gt/lt expressions need to go here.
            if lb_str is not None:
                prep.append(f'greater than ${lb_str}$')
            if ub_str is not None:
                prep.append(f'less than ${ub_str}$')
        if coprime_str:
            prep.append(f'coprime to ${coprime_str}$')
        if dividing_str:
            prep.append(f'dividing ${dividing_str}$')

        # Build the final phrase
        final = symbolic_part + ' ' + noun_phrase
        N = len(prep)
        for i, p in enumerate(prep):
            final += ', '
            if 0 < i == N - 1:
                final += 'and '
            final += p
        return final

    def write_chooser_widget(self):
        return write_int_chooser_widget(
            self.auto_descrip(),
            current_value=self.value
        )


int_chooser_widget_template = jinja2.Template("""
<div class="chooser int_chooser">
    <div class="heading">{{ descrip }}</div>
    <div class="error_display"></div>
    <input class="textfield" type="text" placeholder="Integer" value="{{current_value}}"/>
</div>
""")


def write_int_chooser_widget(descrip, current_value):
    context = {}
    context.update(locals())
    return int_chooser_widget_template.render(context)




# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

import sys


PYODIDE = ('pyodide' in sys.modules)
PFSCSERVER = ('pfsc' in sys.modules)


def adapt(x):
    """
    Adapt an incoming argument if we're running in Pyodide, and the arg is
    given as a JsProxy. Otherwise leave it unchanged.
    """
    if PYODIDE and hasattr(x, 'to_py'):
        x = x.to_py()
    return x



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

"""Utilities for parsing. """

from pfsc_examp.config import config
from pfsc_examp.parse.excep import ParsingExcep


def paren_depth(raw_text):
    """Compute the deepest parenthesis nesting depth in a string."""
    max_depth = 0
    cur_depth = 0
    for c in raw_text:
        if c == "(":
            cur_depth += 1
            max_depth = max(max_depth, cur_depth)
        elif c == ")":
            cur_depth = max(0, cur_depth - 1)
    return max_depth


def check_expr_dims(raw_text, max_len=None, max_depth=None):
    """
    Check that a string is not too long or too deep (in parenthesis nesting),
    for safe parsing as a mathematical expression.

    @param raw_text: the string to be checked.
    @param max_len: maximum allowed length for the string. If `None`, we look
        for a value in the global config.
    @param max_depth: maximum allowed parenthesis depth for the string. If
        `None`, we look for a value in the global config.
    @return: nothing
    @raise: ParsingExcep if given text is too long, or has too deep parenthesis
      nesting.
    """
    if max_len is None:
        max_len = config["MAX_SYMPY_EXPR_LEN"]
    if max_depth is None:
        max_depth = config["MAX_SYMPY_EXPR_DEPTH"]
    if len(raw_text) > max_len:
        raise ParsingExcep("Math expression too long.")
    if paren_depth(raw_text) > max_depth:
        raise ParsingExcep("Math expression too deep.")



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from sympy.core.numbers import Integer
from sympy.logic.boolalg import BooleanAtom
from sympy.ntheory.primetest import isprime
import jinja2

from pfsc_examp.calculate import calculate
from pfsc_examp.canon_arithmeticus import primes_under_100, three_digit_primes
from pfsc_examp.parameters.base import Parameter
from pfsc_examp.contexts import ContextNames
from pfsc_examp.excep import MalformedParamRawValue


class Prime_Param(Parameter):
    """
    Prime rational integer.

    ptype: "Prime"

    Value type: Integer

    Default: int

    Args:
      Required: None.
      Optional:
        odd: boolean, default=False
            True if you require that it be an odd prime
        chooserUpperBound: int
            If positive, do not display primes larger than this in
            the chooser HTML. Note: This is not a *mathematical* constraint;
            you are not saying that the prime actually needs to be bounded
            for any mathematical reason. You are only saying that you want
            to limit the set of primes "on offer". So the reasons for the limit
            are practical ones, relating to computation time, or surveyability
            of displays.
    """

    arg_spec = {
        "OPT": {
            'odd': {
                'type': BooleanAtom,
                'default_raw': False,
            },
            'chooserUpperBound': {
                'type': Integer,
                'default_raw': 0,
            },
        }
    }

    def __init__(self, parent, tex=None,
                 default=None, descrip=None, params=None,
                 args=None, last_raw=None, **other):
        super().__init__(parent, tex, default, descrip, params, args, last_raw)
        self.odd_only = None
        self.chooserUpperBound = None

    def prebuild(self):
        self.odd_only = self.resolved_args['odd']
        self.chooserUpperBound = self.resolved_args['chooserUpperBound']

    def auto_build(self):
        raw = 3 if self.odd_only.value else 2
        return self.build_from_raw(raw)

    def build_from_raw(self, raw):
        n = self.check_int(raw)
        if not calculate(isprime, n):
            raise MalformedParamRawValue('Must be prime.', self)
        if self.odd_only.value and n in [2, -2]:
            raise MalformedParamRawValue('Prime must be odd.', self)
        return Integer(n)

    def auto_descrip(self, include_value=True, editable=True):
        symbolic = self.write_name_and_value(include_value, editable)
        noun = 'prime number'
        if self.parent.context == ContextNames.AlgNT:
            noun = 'rational prime'
        return f"{symbolic} a{'n odd' if self.odd_only.value else ''} {noun}"

    def write_chooser_widget(self):
        # Booleans to control display of widgets:
        show_three_digit_primes = True
        show_free_box = True
        # Prepare our list of primes under 100 (possibly modified):
        pu100 = primes_under_100[1 if self.odd_only.value else 0:]
        if 0 < self.chooserUpperBound.value <= 100:
            show_three_digit_primes = False
            show_free_box = False
            pu100 = filter(lambda p: p <= self.chooserUpperBound.value, pu100)
        # Prepare our list of three-digit primes (possibly modified):
        tdp = None
        if show_three_digit_primes:
            tdp = three_digit_primes[:]
            if 100 < self.chooserUpperBound.value <= 1000:
                show_free_box = False
                tdp = filter(lambda p: p <= self.chooserUpperBound.value, tdp)
        context = {
            'selected_prime': self.value,
            'descrip': self.auto_descrip(),
            'primes_under_100': pu100,
            'three_digit_primes': tdp,
            'show_tdp': show_three_digit_primes,
            'show_free_box': show_free_box,
        }
        return prime_chooser_widget_template.render(context)


prime_chooser_widget_template = jinja2.Template("""
<div class="chooser radio_panel_chooser prime_chooser">
    <div class="heading">{{ descrip }}</div>
    <div class="error_display"></div>
    {# #}
    {# Buttons for the primes under 100 #}
    <div class="radio_panel">
        {% for p in primes_under_100 %}
            {% if p == selected_prime %}
                <div tabindex="0" value="{{ p }}" class="radio_panel_button rpb_selected">{{ p }}</div>
            {% else %}
                <div tabindex="0" value="{{ p }}" class="radio_panel_button">{{ p }}</div>
            {% endif %}
        {% endfor %}
    </div>
    {# #}
    {% if show_tdp %}
        {# Dropdown for the three-digit primes #}
        <select class="dd">
            <option class="dd_null_opt" value="">3-digit primes</option>
            {% for p in three_digit_primes %}
                {% if p == selected_prime %}
                    <option class="dd_opt" value="{{p}}" selected="selected">{{p}}</option>
                {% else %}
                    <option class="dd_opt" value="{{p}}">{{p}}</option>
                {% endif %}
            {% endfor %}
        </select>
    {% endif %}
    {# #}
    {% if show_free_box %}
        {# Text field for anything larger #}
        {% if selected_prime > 1000 %}
            <input class="textfield" type="text" placeholder="Other prime" value="{{selected_prime}}"/>
        {% else %}
            <input class="textfield" type="text" placeholder="Other prime"/>
        {% endif %}
    {% endif %}
</div>
""")



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from pfsc_examp.util import PFSCSERVER
from pfsc_util.imports import from_import


def simple_calc(f, *args, **kwargs):
    return f(*args, **kwargs)


"""
By default, to calculate is simply to apply a function to arguments.
"""
calculate = simple_calc


"""
If we're running in pfsc-server, then we want to use controlled calculation.

EDIT (230119): For now we're disabling this. Server side math jobs are no
longer being used, and now the use of RQ in pise/server unit tests has begun
to cause issues with pickling of certain SymPy types. Probably this old feature
should just be removed.

TODO: Remove support for server-side math jobs.
"""
if PFSCSERVER and False:  # Disabled. See note above.
    calculate = from_import('pfsc_examp.calculate.server_calc', 'calculate')


"""
RQ jobs will not accept a class as the callable. So we need a function we
can call, in order to construct class instances within `calculate`.

E.g.

    f = construct_instance(Poly, x**2 + 1, domain=QQ)

This is only needed when using server_calc, but since we want to write code
that works in any setting, parameter classes should always use it.
"""
construct_instance = simple_calc



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from sympy import AlgebraicField, Integer, latex

from pfsc_examp.calculate import calculate
from pfsc_examp.parameters.base import (
    Parameter,
    write_radio_panel_chooser_widget,
)
from pfsc_examp.excep import MalformedParamRawValue


class PrimeIdeal_Param(Parameter):
    """
    A prime ideal in a number field.

    The chooser HTML offers a radio panel, with which to select one
    of the prime ideals lying above a particular rational prime, in a
    particular number field.

    ptype: "PrimeIdeal"

    Value type: sympy.polys.numberfields.primes.PrimeIdeal

    Default: int
        Zero-based index into the list of PrimeIdeals computed by SymPy.

    Args:
        Required:
            k: param-valued
                Must point to a param widget of `NumberField` type.
            p: int-valued
                The rational prime whose ideal divisors we are to represent.
    """

    arg_spec = {
        "REQ": {
            'k': {
                'type': AlgebraicField,
            },
            'p': {
                'type': Integer,
            }
        }
    }

    def __init__(self, parent, tex=None,
                 default=None, descrip=None, params=None,
                 args=None, last_raw=None, **other):
        super().__init__(parent, tex, default, descrip, params, args, last_raw)
        self.field = None
        self.prime = None
        self.prime_ideals = None
        self.display_values = None
        self.selected_index = None

        self.last_field_value = None
        self.last_prime_value = None

    def prebuild(self):
        K0 = self.last_field_value
        p0 = self.last_prime_value

        self.field = self.resolved_args['k']
        self.prime = self.resolved_args['p']
        K = self.field.value
        p = self.prime.value

        if K != K0 or p != p0:
            # `AlgebraicField.primes_above()` expects a raw `int`, not an `Integer`.
            self.prime_ideals = calculate(K.primes_above, int(p))
            self.display_values = [self.write_display_value_for_prime_ideal(P)
                                   for P in self.prime_ideals]
            self.last_attempted_raw_value = None
            self.last_field_value = K
            self.last_prime_value = p

    def write_display_value_for_prime_ideal(self, P):
        field_gen = self.field.get_field_generator_symbol()
        alpha = P.alpha.poly(x=field_gen).as_expr()
        p = self.prime.value
        return f'({p}, {latex(alpha)})'

    def auto_build(self):
        return self.build_from_raw(0)

    def build_from_raw(self, raw):
        i = self.check_int(raw)
        if not 0 <= i < len(self.prime_ideals):
            raise MalformedParamRawValue(f'Bad index for prime ideal: {i}', self)
        self.selected_index = i
        return self.prime_ideals[i]

    def auto_descrip(self, include_value=True, editable=True):
        display_value = self.display_values[self.selected_index]
        return f'%s a prime ideal over $%s$ in $%s$' % (
            self.write_name_and_value(include_value, editable, display_value),
            self.prime.value, self.field.getTeX(as_name=True)
        )

    def write_chooser_widget(self):
        return write_radio_panel_chooser_widget(
            self.auto_descrip(),
            enumerate([f'${v}$' for v in self.display_values]),
            self.selected_index
        )



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

import traceback
from markupsafe import escape

from displaylang.exceptions import ControlledEvaluationException

from pfsc_examp.contexts import ContextNames
from pfsc_examp.parameters import make_param
from pfsc_examp.parameters.base import Parameter
from pfsc_examp.display import make_disp
from pfsc_examp.excep import ExampError, MalformedParamRawValue
from pfsc_examp.util import adapt
from pfsc_util.imports import from_import


__version__ = "0.25.0"


def make_examp_generator_obj_from_js(info, pane_id):
    """
    This is the factory function for Parameters and ExampDisplays, which is
    intended to be invoked from the JS side, in pfsc-ise.

    Note that we deliberately do not call `to_js()` on the return value, because
    we want a PyProxy on the JS side. When one param or display needs to get
    ahold of another, it will be through `PyodideWidgetStandIn.get_generator()`,
    and will be a "round trip conversion" of a Python object
    (see <https://pyodide.org/en/stable/usage/type-conversions.html#round-trip-conversions>)
    which means this PyProxy will simply be unwrapped to recover the original
    object itself.
    """
    info = adapt(info)
    parent = PyodideWidgetStandIn(info, pane_id)
    wtype = info.get('type')
    if wtype == "PARAM":
        obj = make_param(parent, info)
    elif wtype == "DISP":
        obj = make_disp(parent, info)
    else:
        raise ExampError(f"Invalid widget type {wtype}")
    # NOTE: Tried using `create_proxy()` to ensure persistence on the JS side,
    # like this,
    #   create_proxy = from_import('pyodide', 'create_proxy')
    #   proxy = create_proxy(obj)
    #   return proxy
    # but it didn't work. We still got the,
    #   "This borrowed proxy was automatically destroyed at the end of a function call."
    # error when we tried to use `PyodideWidgetStandIn.get_generator()` more than
    # once on the same widget.
    # This is why the `ExampWidget` class in pfsc-ise offers
    # the `getPyProxyCopy()` method, which returns a *copy* of the proxy, so
    # that the original will not be destroyed.
    # The docs,
    #   https://pyodide.org/en/stable/usage/type-conversions.html#calling-javascript-functions-from-python
    # say that either method should work, but for us only making the copy on the
    # JS side worked. This was with Pyodide v0.19.0.
    # See also:
    #   https://github.com/pyodide/pyodide/issues/1607
    return obj


class ErrCode:
    OK = 0
    UNKNOWN = 1
    UNEXPECTED = 2
    # For historical reasons (original error codes in pfsc-server), we're using
    # the original number 193 for this error:
    MALFORMED_PARAM_RAW_VALUE = 193
    EXAMP_ERROR = 4
    # Again, we want to be sure we're using an error code that doesn't collide
    # with other codes in pfsc-server, so we've set this one to be the same
    # over there.
    CONTROLLED_EVALUATION_EXCEPTION = 279

    # Don't use anything in the range 10,000 - 10,999.
    # This range is reserved for use by mathworker.js, on the client side.
    RESERVED_FOR_MATHWORKER_ON_JS_SIDE_0 = 10000
    RESERVED_FOR_MATHWORKER_ON_JS_SIDE_1 = 10999


def rebuild_examp_generator_from_js(obj, value=None, write_html=False):
    to_js = from_import('pyodide', 'to_js')
    d = {
        'err_lvl': ErrCode.UNKNOWN,
        'err_msg': 'unknown error',
    }
    try:
        obj.build(raw=value)
        html = obj.write_html() if write_html else None
    except MalformedParamRawValue as e:
        d['err_lvl'] = ErrCode.MALFORMED_PARAM_RAW_VALUE
        d['err_msg'] = str(e)
        d['blame_widget_uid'] = e.param.getUid()
        d['trace'] = '\n'.join(traceback.format_stack())
    except ControlledEvaluationException as e:
        d['err_lvl'] = ErrCode.CONTROLLED_EVALUATION_EXCEPTION
        d['err_msg'] = str(e)
        d['blame_widget_uid'] = obj.getUid()
        d['trace'] = '\n'.join(traceback.format_stack())
    except ExampError as e:
        d['err_lvl'] = ErrCode.EXAMP_ERROR
        d['err_msg'] = str(e)
        d['trace'] = '\n'.join(traceback.format_stack())
    except Exception as e:
        d['err_lvl'] = ErrCode.UNEXPECTED
        d['err_msg'] = str(e)
        d['trace'] = '\n'.join(traceback.format_stack())
    else:
        d['err_lvl'] = 0
        d['err_msg'] = ''
        if write_html:
            d['html'] = html
    # Error messages may contain angle brackets (e.g. when reporting expected
    # type for an argument to an allowed callable), and we want these to
    # display correctly in the browser, so we escape:
    d['err_msg'] = escape(d.get('err_msg', ''))
    return to_js(d)


class PyodideWidgetStandIn:
    """
    For execution in Pyodide, plays the role of a Widget, in its capacity as
    the parent object to a Parameter or ExampDisplay.
    """

    def __init__(self, info, pane_id):
        """
        info: (dict) the widget info, as loaded from the anno JSON
        pane_id: (str) the pane Id where we'll look for sibling examp widgets
        """
        self.info = info
        self.pane_id = pane_id

        self.uid = info['uid']
        self.libpath = info['widget_libpath']
        self.dependencies = info['dependencies']
        self.context = info.get('context', ContextNames.Basic)

        self.lp2uid = {
            d["libpath"] : d["uid"] for d in self.dependencies
        }

        hub = from_import('js', 'pfscisehub')
        self.nm = hub.notesManager

    def get_examp_widget(self, lp):
        uid = self.lp2uid[lp]
        return self.nm.getWidget(uid)

    def get_generator(self, lp):
        w = self.get_examp_widget(lp)
        return w.getPyProxyCopy(self.pane_id)

    def getLibpath(self):
        return self.libpath

    def getUid(self):
        return self.uid



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


class ExampError(Exception):
    """Exception base class"""
    pass


class UnknownParameterType(ExampError):
    """Unknown parameter type."""
    pass


class MissingDependencyError(ExampError):
    """Dependency (param or display) could not be found"""
    pass


class MissingParameterError(MissingDependencyError):
    """Parameter could not be found"""
    pass


class MissingDisplayError(MissingDependencyError):
    """Display could not be found"""
    pass


class UnexpectedParamArgError(ExampError):
    """Parameter received an unexpected argument."""
    pass


class UnexpectedParamArgTypeError(ExampError):
    """Parameter received an argument of an unexpected type."""
    pass


class MissingParamArgError(ExampError):
    """Parameter is missing a required arg."""
    pass


class UnresolvedParamArgError(ExampError):
    """Parameter arg value was supplied but could not be resolved."""
    pass


class MissingParamArgType(ExampError):
    """No type was declared for a parameter arg"""
    pass


class MalformedExampImport(ExampError):
    """Import into a display was malformed."""
    pass


class MissingExport(ExampError):
    """Display does not define an expected export."""
    pass


class MalformedInput(ExampError):
    """
    Generic error where user passed input and it was malformed.

    NOTE: If the input was the raw value to a Parameter, then
    `MalformedParamRawValue` should be used instead.
    """
    pass


class MalformedParamRawValue(MalformedInput):
    """The raw value passed to a parameter's build method was malformed."""

    def __init__(self, message, param):
        super().__init__(message)
        self.param = param



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


import dill
import pytest

from sympy import Expr, I

from pfsc_examp.parameters.types.numberfield import NumberField_Param
from pfsc_examp.excep import ExampError
from pfsc_examp.parse.excep import ParsingExcep
from pfsc_examp.parse.polynom import (
    validate_univar_polynom_over_Z
)
from pfsc_examp.parse.algebraic import (
    safe_algebraic_func_str_to_sympy,
)

@pytest.mark.parametrize('given, validated', [
    ["3", {'poly': '3', 'var': 'None'}],
    ["x", {'poly': 'x', 'var': 'x'}],
    ["T", {'poly': 'T', 'var': 'T'}],
    ["2*x + 7", {'poly': '2*x + 7', 'var': 'x'}],
    ["t^2 - t + 18", {'poly': 't**2 - t + 18', 'var': 't'}],
    ["t**2 - t + 18", {'poly': 't**2 - t + 18', 'var': 't'}],
    ["-3*u^17 - 5*u^2 + -2*u + 9", {'poly': '-3*u**17 - 5*u**2 + -2*u + 9', 'var': 'u'}],
])
def test_poly_parse_ok_1(given, validated):
    assert validate_univar_polynom_over_Z(given) == validated


@pytest.mark.parametrize('given', [
    'x^3 + u - 7',
    'x*2',
    'x^-3',
    '3.14*x',
    'x^2 + +x',
    'x^3 + -x',
    'foo',
    'os.system("malicious code")',
])
def test_poly_parse_fail_1(given):
    with pytest.raises(ExampError):
        validate_univar_polynom_over_Z(given)


@pytest.mark.parametrize('given', [
    'x',
    'theta',
    'x^3 + u - 7',
    'x*2',
    'x^-3',
    '3.14*x',
    'x^3 + -x',
    '(2*a^3 + b*c - 7)/(-r + s^(1/2))',
    '(2*a^3 + b*c - 7)/(-theta + s^(x - 1/2))',
    '2 + 3',
    '2 - 3',
    'x + 2',
    '(x - 2)*(y + 3)',
    'a*b + c*d',
    'a/b',
    'b^(-1)',
    '(x^2 - 3*x + 7)^(2*y - 1)',
    'a + b * c ^ d',
    'c ^ d * b + a',
    'a + b + c',
    'p*q*r*s',
    'a - -b',
    'a---b',
])
def test_alg_func_valid(given):
    f = safe_algebraic_func_str_to_sympy(given, max_len=1024, max_depth=32)
    #from sympy import latex
    #print(latex(f))
    assert isinstance(f, Expr)


@pytest.mark.parametrize('given', [
    "sin(x)",
    "x.n.__globals__['__builtins__']['exec']('''import os; os.system(\"echo 'arbitrary code execution...'\")''')",
])
def test_alg_func_invalid(given):
    with pytest.raises(ParsingExcep):
        safe_algebraic_func_str_to_sympy(given)


def test_nf_param_root_idx():
    # When we do not specify the root index, it defaults to -1, which in
    # this case means the cube root of 2 in the second quadrant:
    d = {
        'name': 'k',
        'default': 'x^3 - 2',
    }
    nfp = NumberField_Param(None, **d)
    nfp.build()
    k = nfp.value
    assert abs(k.ext.n(5) - (-0.62996 + 1.0911*I)) < 1e-4

    # Now we specify the real cube root of 2:
    d['args'] = {
        'rootIdx': 0
    }
    nfp = NumberField_Param(None, **d)
    nfp.build()
    k = nfp.value
    assert abs(k.ext.n(5) - 1.2599) < 1e-4


def test_nf_pickle():
    d = {
        'name': 'k',
        'default': 'x^3 - 2',
    }
    nfp = NumberField_Param(None, **d)
    nfp.build()
    k1 = nfp.value
    k2 = dill.loads(dill.dumps(k1))
    assert k1.ext.minpoly == k2.ext.minpoly



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from sympy.core.numbers import Integer, NegativeOne, Zero, One
from sympy.ntheory.factor_ import divisors, divisor_count

from pfsc_examp.calculate import calculate
from pfsc_examp.parameters.base import (
    Parameter,
    write_radio_panel_chooser_widget
)
from pfsc_examp.excep import MalformedParamRawValue


class Divisor_Param(Parameter):
    """
    Represents a rational integer that is supposed to divide another.

    The reason for having a special Parameter class for this (instead of
    just using `Integer_Param`) is to supply the user with a radio panel
    where they are offered all the possible divisors. Remember that the
    purpose of a Parameter class is not just to represent a particular
    type of mathematical object, but to represent such an object _plus
    a way of choosing its values_.

    ptype: "Divisor"

    Value type: Integer

    Default: int

    Args:
        Required:
            dividing: int-valued
                The number of which this is a divisor.
        Optional:
            sign: {-1, 0, 1}, default=1
                1 if you want only positive divisors
                -1 if you want only negative divisors
                0 if you allow both positive and negative divisors

    """

    arg_spec = {
        "REQ": {
            'dividing': {
                'type': Integer,
            },
        },
        "OPT": {
            'sign': {
                'type': (One, Zero, NegativeOne,),
                'default_cooked': Integer(1),
            },
        }
    }

    def __init__(self, parent, tex=None,
                 default=None, descrip=None, params=None,
                 args=None, last_raw=None, **other):
        super().__init__(parent, tex, default, descrip, params, args, last_raw)
        self.divisors = []
        self.dividing = None
        self.sign = None

    def prebuild(self):
        # Store args as attributes
        self.dividing = self.resolved_args['dividing']
        self.sign = self.resolved_args['sign']
        # Compute all the divisors.
        n = self.dividing.value
        c = calculate(divisor_count, n)
        if c > 200:  # FIXME: make such upper bounds configurable.
            msg = f'Parameter {self.getLibpath()} has too many notes, er divisors.'
            raise MalformedParamRawValue(msg, self)
        self.divisors = calculate(divisors, n)
        s = self.sign.value
        if s == 0:
            self.divisors = list(reversed([-d for d in self.divisors])) + self.divisors
        elif s == -1:
            self.divisors = list(reversed([-d for d in self.divisors]))

    def auto_build(self):
        s = self.sign.value
        d = s if s != 0 else 1
        return self.build_from_raw(d)

    def build_from_raw(self, raw):
        d = self.check_int(raw)
        n = self.dividing.value
        if not n % d == 0:
            raise MalformedParamRawValue(f'Must divide ${n}$', self)
        return Integer(d)

    def auto_descrip(self, include_value=True, editable=True):
        return '%s a divisor of $%s$' % (
            self.write_name_and_value(include_value, editable),
            self.dividing
        )

    def write_chooser_widget(self):
        return write_radio_panel_chooser_widget(
            self.auto_descrip(), zip(self.divisors, self.divisors),
            self.value
        )



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from pfsc_examp.util import (
    PYODIDE,
    PFSCSERVER,
)
from pfsc_util.imports import from_import

# Default values
config = {
    "MAX_SYMPY_EXPR_LEN": 1024,
    "MAX_SYMPY_EXPR_DEPTH": 32,
    "MAX_DISPLAY_BUILD_LEN": 4096,
    "MAX_DISPLAY_BUILD_DEPTH": 32,
}

# Override default values, depending on context
if PFSCSERVER:
    has_app_context, current_app = from_import('flask', ['has_app_context', 'current_app'])
    if has_app_context():
        config = current_app.config
elif PYODIDE:
    pfscExampConfig = from_import('js', 'pfscExampConfig')
    config = pfscExampConfig.to_py()



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

import jinja2
from markupsafe import Markup

from sympy import S as SS
from sympy import Symbol, latex

from pfsc_examp.parse.algebraic import safe_algebraic_func_str_to_sympy
from pfsc_examp.parse.excep import ParsingExcep
from pfsc_examp.excep import (
    MissingParamArgError,
    MissingParameterError,
    MalformedParamRawValue,
    UnexpectedParamArgError,
    UnexpectedParamArgTypeError,
    UnresolvedParamArgError,
    MissingParamArgType,
)
from pfsc_examp.util import adapt
from pfsc_util.imports import from_import


class Valued:
    """
    Abstract base class for objects with a `.value` proeprty.
    """

    @property
    def value(self):
        return self._value


class NonParamValue(Valued):
    """
    Simple wrapper class for values that are not dependent on `Parameter`s.
    The internal `self._value` should be an instance of some SymPy class.
    """

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return latex(self._value)


class ParametricValued(Valued):
    """
    Superclass of any `Valued` that is parametric, i.e. whose value depends
    on the value(s) of one or more Parameters.
    """

    def get_libpaths(self):
        """
        @return: iterable of all libpaths of parameters involved in this object.
        """
        raise NotImplementedError


class ParametricExpression(ParametricValued):
    """
    For SymPy expressions containing symbols that stand for Parameters.
    """

    def __init__(self, expr, params, require_saturated=False):
        """
        @param expr: a SymPy Expr
        @param params: dict mapping names (strings) to Parameter instances.
          Each Symbol occuring in `expr` whose `.name` is a key in this dict
          will be replaced by the value of that Parameter, when the value of
          this ParametricExpression is requested.
        @param require_saturated: If True, and if not every symbol name occurs
          as a parameter name, then raise an exception.
        """
        self.expr = expr
        self.params = params

        self.symbols = {a for a in expr.atoms() if isinstance(a, Symbol)}
        self.names = [a.name for a in self.symbols]

        if require_saturated:
            for a in self.names:
                if a not in self.params:
                    msg = f'Expression contains symbol `{a}` corresp. to no parameter'
                    msg += ', but is required to be saturated.'
                    raise MissingParameterError(msg)

        self.orig_names = {s: self.params[s.name].getTeX(as_name=True)
                           for s in self.symbols if s.name in self.params}

    def __str__(self):
        return latex(self.expr, symbol_names=self.orig_names)

    @property
    def value(self):
        v = [(name, param.value) for name, param in self.params.items()]
        return self.expr.subs(v, simultaneous=True)

    def get_libpaths(self):
        return [param.getLibpath() for param in self.params.values()]


class Parameter(ParametricValued):
    """
    The abstract base class for all parameter types.
    """

    def __init__(self, parent, tex,
                 default=None, descrip=None, params=None,
                 args=None, last_raw=None):
        """
        :param parent: reference to a governing object. Could be a
          PyodideWidgetStandIn when operating in Pyodide, or a ParamWidget when
          operating in pfsc-server.

        :param tex: string to be used in TeX representations.

        :param default: some specification of a default value for this parameter.
          May be of various types; it is up to the subclass to interpret.
          If specified, will be passed through the same method (build_from_raw)
          as any user-selected value that may come down from the client side.

          May be left unspecified (None), and then the subclass's auto_build
          method will be engaged at build time.

        :param descrip: optional string giving a textual description of this
          parameter, for the benefit of users; however, it is more common to
          let the description instead be generated by the subclass's
          auto_descrip method.

        :param params: dict mapping local names to libpaths of other parameters.
          The names appearing here may then occur within mathematical expressions
          appearing as values in the `args` dict (see below). The value of such
          expressions will then be obtained by substituting the values of these
          parameters.

          NOTE: As a convenience, if you import a param under the same name as
          one of the args (see below) this will set the value of the parameter
          as the value of the arg, unless you override it by defining the
          arg to have some other value.

        :param args: dict giving any arguments, required and/or optional, for
          defining this parameter. Keys are argument names (prescribed by the
          particular subclass); values are any raw JSON type that in some way
          specifies a value (to be interpreted by the particular subclass's
          `parse_raw_arg` method).

          Shortcut option: you may omit an arg definition if you have imported
          a parameter under the exact same name. By doing so, you imply that
          you want the value of that parameter as the value of this arg.

        :param last_raw: useful when rebuilding a parameter that has already
          been in use, in the ISE. You can pass the most recent raw value that
          this parameter has taken on, and it will be recorded as such here. In
          particular, it will take precedence over `self.default` when we build.
        """
        self.parent = parent
        self.tex = tex
        self.default = default
        self.last_attempted_raw_value = last_raw

        self.descrip = descrip
        self.given_args = args or {}
        # Dict in which param name points to Parameter instance:
        self.resolved_params = {k: self.parent.get_generator(v)
                                for k, v in (params or {}).items()}
        # Dict in which arg name points to Valued instance:
        self.resolved_args = self.resolve_args(self.given_args)
        # Initialize value to None, so we know whether the parameter has been
        # built yet.
        self._value = None

    def resolve_args(self, args):
        """
        @param args: the raw `args` dict defined in a parameter widget's data.
        @return: dict mapping arg names to instances of `Valued`.
        """
        resolved = {}
        for mode in ["REQ", "OPT"]:
            items = self.arg_spec.get(mode, {}).items()
            for name, spec in items:
                value = None
                if name in args:
                    raw = args[name]
                elif name in self.resolved_params:
                    # Shortcut rule: if you didn't define an arg, but you did
                    # import a param under the exact same name, then you imply
                    # that you want this param as the value of this arg.
                    value = self.resolved_params[name]
                elif mode == "REQ":
                    raise MissingParamArgError(f"Required parameter arg `{name}` not supplied.")
                elif 'default_raw' in spec:
                    raw = spec['default_raw']
                elif 'default_cooked' in spec:
                    value = NonParamValue(spec['default_cooked'])
                else:
                    continue
                if value is None:
                    value = self.parse_raw_arg(name, raw)
                if not isinstance(value, Valued):
                    raise UnresolvedParamArgError(name)
                resolved[name] = value
        return resolved

    def check_value_types(self):
        """
        Check that each resolved arg has value of the required type.
        If not, raise an exception.

        This is invoked before `self.prebuild()`.
        """
        for mode in ["REQ", "OPT"]:
            items = self.arg_spec.get(mode, {}).items()
            for name, spec in items:
                if name in self.resolved_args:
                    a = self.resolved_args[name]
                    t = spec.get('type')
                    if t is None:
                        raise MissingParamArgType(name)
                    if not isinstance(a.value, t):
                        raise UnexpectedParamArgTypeError(f'Arg `{name}` has wrong type.')

    def get_arg_value(self, name, default=None):
        """
        Retrieve the value of a resolved arg, or return a default value if the
        arg was not defined.

        @param name: (str) the name of the arg
        @param default: (any) value to return if the name is not found
        @return: the value of the arg, or the default value
        """
        if not self.has_arg(name):
            return default
        return self.resolved_args[name].value

    def has_arg(self, name):
        return name in self.resolved_args

    def get_arg_mode_and_expected_type(self, name):
        """
        @param name: name of an arg defined in `self.arg_spec`
        @return: pair (m, t), where m is the arg mode "REQ" or "OPT",
          and t is a tuple listing all the allowed types for the value
          of the arg.
        """
        for mode in ["REQ", "OPT"]:
            t = self.arg_spec.get(mode, {}).get(name, {}).get('type')
            if t:
                if not isinstance(t, tuple):
                    t = (t,)
                return mode, t
        raise UnexpectedParamArgError

    def check_int(self, raw):
        try:
            n = int(raw)
        except Exception:
            raise MalformedParamRawValue(raw, self)
        return n

    @property
    def value(self):
        if self._value is None:
            self.build()
        return self._value

    def getTeX(self, as_name=True):
        return self.tex if as_name else latex(self.value)

    def __str__(self):
        return self.getTeX()

    def getUid(self):
        return self.parent.getUid()

    def getLibpath(self):
        return self.parent.getLibpath()

    def get_libpaths(self):
        return [self.getLibpath()]

    def write_name_and_value(self, include_value=True, editable=True, display_value=None):
        """
        A utility for use by subclasses' auto_descrip methods. Achieves a common
        pattern in which the value of the parameter is rendered according to the
        keyword args, which are the same as for the auto_descrip method, except that
        the additional arg display_value may be provided if something other than
        self.value should be shown.
        """
        s = f'${self}$'
        if include_value:
            if display_value is None:
                display_value = self.value
            if editable:
                v = f'<span class="param_val">&nbsp;$= {display_value}$</span>'
            else:
                v = f'&nbsp;$= {display_value}$'
            s += Markup(v)
        return s

    def write_html(self):
        """
        Note: It's important that this be a normal method, not a @property
        method. In Pyodide (as of v0.19.0), it seems that exceptions raised
        while executing property methods are silenced. But we need to see any
        exceptions, so we can debug.
        """
        return self.write_chooser_widget()

    def build(self, raw=None):
        """
        This is where `self._value` gets defined. This should be an instance of
        some class defined in SymPy. This includes things like a `Poly` or an
        `AlgebraicField`, but also basic types like `Integer`, `Rational`,
        `Float`, and `BooleanAtom` (for True and False). In other words,
        even Python primitives should be transformed into appropriate SymPy
        objects.

        If `self.default` has been provided, it should be some primitive
        value that somehow specifies the desired value for this parameter. For
        example, it might be an int, or a string describing a polynomial.
        It has to be something that could be written into a widget's data (i.e.
        a valid JSON type), but now it is to be transformed into an internal
        representation. It is passed to `self.build_from_raw()` in order to
        achieve this transformation.

        If `self.default` was not provided, then we "auto build", choosing a
        value perhaps based on optional args in the arg_sepc, perhaps based on
        prior parameters on which we depend, or perhaps based on nothing.
        Subclasses decide how this works in their `self.auto_build()` method.
        One common pattern is to pick a raw value, and pass it to
        `self.build_from_raw()`.

        There are optional prebuild and postbuild hooks that subclasses may
        override.
        """
        raw = adapt(raw)
        if raw is not None:
            self.last_attempted_raw_value = raw
        self._value = None

        self.check_value_types()
        self.prebuild()

        if raw is not None:
            # When attempting to build based on a user-supplied raw value, we do
            # NOT catch exceptions. The user has to know that their value was
            # attempted, and failed.
            self._value = self.build_from_raw(raw)
        else:
            # When attempting to build based on a previously attempted value,
            # or an author-supplied default, we DO catch exceptions.
            # In the case of previously attempted values, the user has not
            # actively entered a value, so should be happy to have the chooser
            # switch back to anything that works.
            # In the case of author-supplied defaults, we want authors to be
            # free to make defaults that work with the *initial* values of other
            # parameters, but not necessarily with other values those params
            # may take on later.
            raw = self.last_attempted_raw_value
            if raw is None:
                raw = self.default
            if raw is not None:
                try:
                    self._value = self.build_from_raw(raw)
                except MalformedParamRawValue:
                    pass
            if self._value is None:
                self._value = self.auto_build()

        self.postbuild()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # SUBCLASSES MAY OVERRIDE

    """
    `arg_spec` is where subclasses define their required and optional args.
    E.g. might look something like:
    
        arg_spec = {
            "REQ": {
                'dividing': {
                    'type': Integer,
                },
            },
            "OPT": {
                'sign': {
                    'type': (One, Zero, NegativeOne,),
                    'default_cooked': Integer(1),
                },
            }
        }
    
    In particular, each arg should define a 'type', whose value should be
    either a single SymPy type, or else a tuple thereof (serving as a
    disjunctive type, i.e. input may define any of these).
    
    Note that the 'type' is not the raw, JSON type of the input, but the
    eventual type of the internal SymPy object the raw input specifies in one
    way or another (it's up to the particular parameter subclass to interpret;
    see `parse_raw_arg()` method).
    
    Optional args may define a 'default_raw', or `default_cooked`, which
    supplies a value in case the user did not supply one. Neither is required.
    
    A 'default_raw' value will be passed through the same parsing process as a
    raw value supplied by the user, while a 'default_cooked' value is assumed
    to already be of the internal SymPy type resulting from that process, and
    is accepted as is.
    
    All types will be checked at the start of the `build()` process, and an
    exception raised if any check fails. 
    """
    arg_spec = {}

    def parse_raw_arg(self, name, raw):
        """
        Transform a raw parameter argument value (a JSON type) into its
        internal representation as some SymPy type.

        Subclasses that expect any of their args to be (even optionally) passed
        via raw JSON types (i.e. not by reference to other parameters) may need
        to override this method.

        A common pattern for subclasses should be to call this method (as a
        superclass method), and iff it returns `None`, then consider other
        possible cases.

        This method will return a `Valued` object on any raw arg of type
        `bool`, `int`, or `float`, and on _some_ raw args of type `str` (see
        below). Otherwise it returns `None`.

        This method will turn raw args of types `bool`, `int`, and `float` into
        their corresponding SymPy types, and wrap these in a `NonParamValue`.

        Raw args of type `str` it will attempt to parse as algebraic
        expressions. This includes anything as simple as a single symbol
        (like 'x' or 'theta') up to any expression built using symbol names,
        numbers (int and float), and the binary operations +, -, *, /, ^, **,
        where ^ is a synonym for **.

        When a `str` parses successfully as a Sympy `Expr`, we wrap it in a
        `ParametricExpression`.

        When a `str` does not parse successfully as a Sympy `Expr`, this method
        returns `None`.

        WARNING
        =======

        EDIT: This warning was originally written before we used a custom fork
        of SymPy. We now use a fork in which `parse_expr()` no longer calls
        `eval()`. Nevertheless, our policy is still to use custom parsers
        instead of `S` or `sympify`.

            Subclasses that do wish to do their own parsing
            MUST NOT use Sympy's `S` or `sympify` functions to parse strings!

            Similarly, DO NOT pass strings directly to constructors like `Poly`,
            which just uses `sympify` internally.

            Consider, for example,

            >>> S("x.n.__globals__['__builtins__']['exec']('''import os; os.system(\"echo 'arbitrary code execution...'\")''')")
            arbitrary code execution...

            (Explanation: `sympify` (which is called through `S`) turns `x` into a
            `Symbol`. The `Symbol` class has a method called `n`. Every method in
            Python has a `__globals__` attribute, through which all built-ins (incl
            `exec`) can be reached.)

            Instead, use the parsers defined under `pfsc.examp.parameters.parse`.
            If the parser you need isn't defined there yet, please consider writing
            it and contributing via pull request.

        @param name: the name of the argument.
            Note: Although `name` is not used in this method, it may well be used in
            subclass overrides of this method.
        @param raw: the raw value of the argument.

        @return: an instance of `Valued`, or `None`
        """
        if isinstance(raw, (bool, int, float)):
            return NonParamValue(SS(raw))

        if isinstance(raw, str):
            try:
                expr = safe_algebraic_func_str_to_sympy(raw)
            except ParsingExcep:
                return None
            else:
                return ParametricExpression(expr, self.resolved_params)

        return None

    def prebuild(self):
        """
        Subclasses may assume that, when `prebuild` is called:
            * all prior parameters on which this one depends have already been
              built (i.e. they have defined their `self._value`).
            * all resolved args for this parameter have had their values
              type-checked.

        This is a good place to compute any values that may be used repeatedly
        in the "MUST OVERRIDE" methods below.

        It is also common to grab resolved args and store them under instance
        attributes here, so that they can be more easily referenced.
        Best practice is to store the args themselves, not their values.
        Then, where values are needed, just use `.value`.
        """
        pass

    def postbuild(self):
        pass

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # SUBCLASSES MUST OVERRIDE

    def auto_build(self):
        """
        Choose a value for this parameter. See doctext for self.build()

        :return: the value
        """
        raise NotImplementedError

    def build_from_raw(self, raw):
        """
        Build a value for this parameter from the given raw description.
        See doctext for self.build()

        :param raw: string/numerical description/specification of the desired
          parameter value
        :return: the value
        """
        raise NotImplementedError

    def auto_descrip(self, include_value=True, editable=True):
        """
        Write a string giving a verbal/symbolic description of this parameter.
        Should honor `self.parent.context` when choosing phraseology.

        :param include_value: e.g. "$p$ a prime" if False,
          but "$p = 7$ a prime" if True.
        :param editable: only meaningful if `include_value` is True.
          In that case, `editable` True should yield e.g.

            "$p$ <span class="param_val">&nbsp;$= 7$</span> a prime"

          so that it's easy to edit the value on the client-side.

        :return: the description string

        See Also: write_name_and_value()
        """
        raise NotImplementedError

    def write_chooser_widget(self):
        """
        Write HTML to describe the client-side chooser widget with which users
        may choose new values for this parameter.

        This is not to be invoked until after `self.build()`.

        :return: HTML string
        """
        raise NotImplementedError

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def write_radio_panel_chooser_widget(descrip, code_and_text, selected=None):
    """
    :param descrip: the parameter's description (e.g. '$\\mathfrak{p}$ a prime ideal')
    :param code_and_text: list of pairs (c, t) in which t is the text that should
      appear on the button, e.g. '$(13, x^2 + 3 x + 1)$' and c the code that
      should be passed to the server when this option is selected,
      e.g. '0'. They can be strings, or anything else that formats to the
      desired string.
    :param selected: the code of the selected button
    :return: HTML string for the chooser widget
    """
    context = {}
    context.update(locals())
    return radio_panel_chooser_widget_template.render(context)


radio_panel_chooser_widget_template = jinja2.Template("""
<div class="chooser radio_panel_chooser">
    <div class="heading">{{ descrip }}</div>
    <div class="error_display"></div>
    <div class="radio_panel">
        {% for c, t in code_and_text %}
            {% if c == selected %}
                <div tabindex="0" value="{{ c }}" display="{{ t }}" class="radio_panel_button rpb_selected">{{ t }}</div>
            {% else %}
                <div tabindex="0" value="{{ c }}" display="{{ t }}" class="radio_panel_button">{{ t }}</div>
            {% endif %}
        {% endfor %}
    </div>
</div>
""")



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

"""
This module is where we define the set of allowed callables from SymPy.
The set is expected to grow over time, in response to user demand.
"""

from sympy import (
    AlgebraicField, AlgebraicNumber, binomial, Expr, factorint,
    Integer, isprime, latex,
    Matrix, mod_inverse, Poly, primerange, QQ, S, Symbol, symbols,
)
from sympy.core.evalf import EvalfMixin
from sympy.logic.boolalg import Boolean
from sympy.polys.domains.domain import Domain
from sympy.polys.matrices import DomainMatrix
from sympy.polys.numberfields.modules import (
    ModuleElement, PowerBasis, Submodule
)
from sympy.polys.numberfields.primes import PrimeIdeal
from sympy.polys.polyclasses import ANP

from typing import List, Optional as o, Union as u, Iterable

from displaylang.allow import (
    ArgSpec as a,
    AllowedCallable as c,
    StrPermType as s,
    Tail as t
)


Bool = u[Boolean, bool]
Int = u[Integer, int]
iExpr = u[Expr, int]
fExpr = u[Expr, float]
fiExpr = u[Expr, float, int]


sympy_available_constants = {
    'QQ': QQ,
}

# We need to override some of the built-ins defined in the `displaylang`
# library, to allow them to accept not just `int` but also `Integer`.
generalized_builtin_allowed_callables = [
    c(range, [t(Int)]),
    c(max, [t(fiExpr)]),
    c(min, [t(fiExpr)]),
]

# The "basic" callables are those that will be present in the namespace for all
# display code evaluations, under their `.__name__`, before defining any imported
# names from any other displays or parameters.
sympy_basic_allowed_callables = [

    # Note: We're supplying the `S` function only to provide an easy way to
    # construct a SymPy `Integer`, `Rational`, or `Float`.
    # If you want a `Symbol`, you should use either `symbols()` or `Symbol()`.
    # If we allowed `S` for this purpose, users would find it confusing, since
    # `beta`, `gamma`, `zeta` would be exceptions, being recognized instead as
    # special functions.
    c(S, [u[int, float]], name="S"),

    c(binomial, [Int, Int]),
    c(factorint, [Int]),
    c(isprime, [Int]),
    c(latex, [[Expr], [Matrix]], {'order': s.CNAME}, incomplete=True),
    c(Matrix, [List[List[fiExpr]]], name="Matrix"),
    c(mod_inverse, [Int, Int]),
    c(Poly, [Expr, t(Symbol)], {'modulus': Int}, incomplete=True),
    c(primerange, [Int]),
    c(Symbol, [s.UWORD]),
    c(symbols, [s.UWORD_CDL]),
]

# These are callables not added to the namespace basis. Instead, we expect you
# to access them through other objects you define or import from other displays
# or parameters.
sympy_other_allowed_callables = [
    c(AlgebraicField.integral_basis, [], {'fmt': s.CNAME}, method_of=AlgebraicField),
    c(AlgebraicField.primes_above, [int], method_of=AlgebraicField),
    c(AlgebraicField.new, [List[iExpr]], method_of=AlgebraicField),
    c(AlgebraicField.__call__, [List[iExpr]], method_of=AlgebraicField),
    c(AlgebraicField.to_sympy, [ANP], method_of=AlgebraicField),
    c(AlgebraicField.to_alg_num, [ANP], method_of=AlgebraicField),
    c(Domain.cyclotomic_field, [Int],
      {'ss': bool, 'alias': s.UWORD, 'gen': Symbol, 'root_index': int},
      method_of=Domain),
    # Allows calling e.n() on any Expr e:
    c(EvalfMixin.evalf, [Int], method_of=EvalfMixin),
    c(Expr.expand, [], method_of=Expr),
    c(Expr.subs, [
        [fiExpr, fiExpr],
        [Iterable]
    ], {'simultaneous': Bool}, method_of=Expr),
    c(Matrix.flat, [], method_of=Matrix),
    c(Matrix.transpose, [], method_of=Matrix),
    c(Poly.as_expr, [t(Symbol)], method_of=Poly),
    c(PowerBasis.__call__, [
        [DomainMatrix],
        [int]
    ], {'denom': int}, method_of=PowerBasis),
    c(PrimeIdeal.reduce_element, [ModuleElement], method_of=PrimeIdeal),
    c(PrimeIdeal.reduce_ANP, [ANP], method_of=PrimeIdeal),
    c(PrimeIdeal.reduce_alg_num, [AlgebraicNumber], method_of=PrimeIdeal),
    c(Submodule.__call__, [
        [DomainMatrix],
        [int]
    ], {'denom': int}, method_of=Submodule),
]



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from pfsc_examp.excep import UnknownParameterType
from .types import *


class ParamTypes:
    Integer = "Integer"
    Prime = "Prime"
    Divisor = "Divisor"
    PrimRes = "PrimRes"
    NumberField = "NumberField"
    PrimeIdeal = "PrimeIdeal"


PARAM_TYPE_TO_CLASS = {
    ParamTypes.Integer:     Integer_Param,
    ParamTypes.Prime:       Prime_Param,
    ParamTypes.Divisor:     Divisor_Param,
    ParamTypes.PrimRes:     PrimRes_Param,
    ParamTypes.NumberField: NumberField_Param,
    ParamTypes.PrimeIdeal:  PrimeIdeal_Param,
}


def make_param(parent, info):
    """
    Construct an instance of a Parameter subclass.

    @param parent: the parent object for the instance
    @param info: (dict) the info dict from a param widget definition

    @raises: UnknownParameterType if info['ptype'] is not a known parameter type name.
    @returns: a Parameter subclass instance
    """
    ptype = info.get('ptype')
    PClass = PARAM_TYPE_TO_CLASS.get(ptype)
    if not PClass:
        raise UnknownParameterType(ptype)
    return PClass(parent, **info)



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from lark import Lark, Transformer
from lark.exceptions import LarkError, VisitError

from pfsc_examp.excep import ExampError, MalformedInput
from pfsc_examp.parse.excep import ParsingExcep

univar_polynom_over_Z_grammar = r"""
    poly: term (PM term)*
    PM: "+"|"-"
    ?term: var_term | coeff
    var_term: (coeff "*")? var (("^"|"**") power)?
    coeff: [PM] INT
    var: LETTER
    power: INT

    %import common.LETTER
    %import common.INT
    %import common.WS
    %ignore WS
"""

univar_polynom_over_Z_parser = Lark(
    univar_polynom_over_Z_grammar, start='poly', lexer='standard')

class UnivarPolynomOverZTransformer(Transformer):

    def __init__(self):
        self.var_letter = None

    def coeff(self, items):
        n = items[-1]
        # Regularize (e.g. 002 -> 2)
        n = int(n)
        s = str(n)
        if len(items) == 2:
            sign = items[0]
            assert sign in "+-"
            s = sign + s
        return s

    def var(self, items):
        letter = items[0]
        if self.var_letter is not None and letter != self.var_letter:
            msg = 'Univariate polynomial must only name one variable.'
            raise MalformedInput(msg)
        self.var_letter = letter
        return letter

    def power(self, items):
        # Validated polynomial expressions are intended for consumption by
        # the SymPy library, so exponentiation is given by `**`.
        return f'**{int(items[0])}'

    def var_term(self, items):
        k = len(items)
        if k == 1:
            return items[0]
        elif k == 2:
            a, b = items
            # Powers include their initial "**", but coeffs do NOT include their trailing "*".
            # This is important, so that we can have terms that are pure coeffs.
            if b.startswith("**"):
                return f'{a}{b}'
            else:
                return f'{a}*{b}'
        elif k == 3:
            c, v, p = items
            return f'{c}*{v}{p}'

    def poly(self, items):
        return ' '.join(items)

def validate_univar_polynom_over_Z(raw_text):
    """
    :param raw_text: a string which _is supposed to_ define a univariate polynomial
      with integer coefficients.
    :return: dict {
      poly: a string that _does_ define a univariate polynomial with int coeffs,
      var: the variable (str) or None if constant polynomial
    }
    :raises: ParsingExcep if anything goes wrong during parsing or transforming.
    """
    try:
        ast = univar_polynom_over_Z_parser.parse(raw_text)
        transformer = UnivarPolynomOverZTransformer()
        validated = transformer.transform(ast)
    except VisitError as v:
        # Lark traps our `ExampError`s, but we want to see them.
        if isinstance(v.orig_exc, ExampError):
            raise v.orig_exc from v
        else:
            raise ParsingExcep(f'Parsing error: {v}') from v
    except LarkError as e:
        raise ParsingExcep(f'Parsing error: {e}') from e
    return {
        'poly': validated,
        'var': str(transformer.var_letter),
    }



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

"""Parsing for algebraic functions. """

from sympy import Add, Mul, Pow, Symbol, Integer, Float
from lark import Lark, Transformer
from lark.exceptions import LarkError

from pfsc_examp.parse.util import check_expr_dims
from pfsc_examp.parse.excep import ParsingExcep


algebraic_grammar = r"""
    ?expr: add | mul | pow | atomic
    add: expr (plus | minus) expr
    mul: factor (times | dividedby) factor
    ?factor: mul | pow | atomic
    pow: atomic ("^"|"**") atomic
    ?atomic: "(" expr ")" | atom
    ?atom: pos | neg
    neg: "-" atom
    ?pos: var | int | float
    plus: "+"
    minus: "-"
    times: "*"
    dividedby: "/"
    var: CNAME
    int: INT
    float : FLOAT

    %import common.CNAME
    %import common.INT
    %import common.FLOAT
    %import common.WS
    %ignore WS
"""


algebraic_func_parser = Lark(
    algebraic_grammar, start='expr', lexer='standard')


class ExprBuilder(Transformer):

    def add(self, items):
        a, op, b = items
        if op.data == 'minus':
            b = -b
        return Add(a, b)

    def mul(self, items):
        a, op, b = items
        if op.data == 'dividedby':
            b = Pow(b, -1)
        return Mul(a, b)

    def pow(self, items):
        a, b = items
        return Pow(a, b)

    def neg(self, items):
        return -items[0]

    def var(self, items):
        return Symbol(items[0])

    def int(self, items):
        return Integer(items[0])

    def float(self, items):
        return Float(items[0])


def safe_algebraic_func_str_to_sympy(raw_text, max_len=None, max_depth=None):
    """
    Safely convert a string defining an algebraic function into a SymPy expression.

    @param raw_text: the string supposedly defining the algebraic function.
    @param max_len: maximum allowed length for the string. If `None`, we look
        for a value in the global config.
    @param max_depth: maximum allowed parenthesis depth for the string. If
        `None`, we look for a value in the global config.
    @return: SymPy Expr
    @raise: ParsingExcep if given text is too long, or has too deep parenthesis
      nesting (important to avoid exceeding recursion stack max depth), or if
      it simply doesn't parse as a valid algebraic function.
    """
    check_expr_dims(raw_text, max_len=max_len, max_depth=max_depth)
    try:
        tree = algebraic_func_parser.parse(raw_text)
    except LarkError as e:
        raise ParsingExcep("Expression not a valid algebraic function.\n" + str(e))
    builder = ExprBuilder()
    expr = builder.transform(tree)
    return expr



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

import re

from sympy import (
    AlgebraicNumber, QQ, Poly, cyclotomic_poly, Symbol, latex, CRootOf, Integer
)
import jinja2
from lark.exceptions import LarkError

from pfsc_examp.calculate import calculate, construct_instance
from pfsc_examp.parameters.base import Parameter
from pfsc_examp.parse.polynom import validate_univar_polynom_over_Z
from pfsc_examp.parse.excep import ParsingExcep
from pfsc_examp.excep import MalformedParamRawValue


CYC_POLY_RAW_PATTERN = re.compile(r'^cyc(lotomic_poly)?\((\d+)\)$')

class NumberField_Param(Parameter):
    """
    Number Field over the rationals.

    ptype: "NumberField"

    Value type: AlgebraicField

    Default: str
        accepts EITHER:
        - `cyc(n)` or `cyclotomic_poly(n)` for the nth cyclotomic field, OR
        - a string giving any irreducible univariate polynomial over Z

    Args:
        Required: None.
        Optional:
            gen: str
                Symbol name to use for a generator of the field.
                For example, 'theta'.
                If supplied, it will be included in the description of the
                field; otherwise, we only describe the field as a quotient
                mod a polynomial.
            var: str
                Symbol name to use for the variable in the field's defining,
                irreducible polynomial.
                If not supplied, we use `x` for `cyc(n)`, or the
                given variable for polynomials given explicitly.
            rootIdx: int
                Root index, indicating which root of the minimal polynomial to
                use as primitive element for the field. This is passed to
                SymPy's `CRootOf`. Real roots come first, in increasing order,
                followed by complex roots, which are sorted first by real part
                (increasing) and then by imaginary part (increasing).
                If not supplied, we use -1.
    """

    arg_spec = {
        "OPT": {
            'gen': {
                'type': Symbol,
            },
            'var': {
                'type': Symbol,
            },
            'rootIdx': {
                'type': Integer,
                'default_raw': -1,
            }
        }
    }

    def __init__(self, parent, tex=None,
                 default=None, descrip=None, params=None,
                 args=None, last_raw=None, **other):
        super().__init__(parent, tex, default, descrip, params, args, last_raw)
        self.explicit_cyclo = False
        self.cyc_num = None
        self.raw = None
        self.validated = None

    def get_poly(self):
        return self.value.ext.root.minpoly

    def get_field_generator_symbol(self):
        gen = self.resolved_args.get('gen')
        return gen.value if gen is not None else None

    def write_polynomial(self):
        if self.explicit_cyclo:
            pol = fr'\Phi_{self.cyc_num}({self.get_poly().gen})'
        else:
            pol = latex(self.get_poly().as_expr())
        return pol

    def write_as_quotient(self):
        var = self.get_poly().gen
        pol = self.write_polynomial()
        return fr'\mathbb{{Q}}[{var}]/({pol})'

    def auto_descrip(self, include_value=True, editable=True):
        display_value = self.write_as_quotient()
        gen = self.resolved_args.get('gen')
        if gen is not None:
            display_value = fr'\mathbb{{Q}}({gen}) \cong ' + display_value
        name_part = self.write_name_and_value(include_value, editable, display_value=display_value)
        noun_phrase = r'a number field over $\mathbb{Q}$'
        return f'{name_part} {noun_phrase}'

    def auto_build(self):
        raw = 'cyc(2)'
        return self.build_from_raw(raw)

    def build_from_raw(self, raw):
        """
        :return: AlgebraicField
        """
        raw = str(raw)  # Ensure that we have a string
        M = CYC_POLY_RAW_PATTERN.match(raw)
        if M:
            self.explicit_cyclo = True
            self.cyc_num = int(M.group(2))
            validated = raw
            sym = self.resolved_args.get('var', Symbol('x'))
            poly_expr = calculate(cyclotomic_poly, self.cyc_num, sym)
            poly_value = calculate(construct_instance, Poly, poly_expr)
        else:
            # Use `calculate` here in case the description is too long.
            try:
                v = calculate(validate_univar_polynom_over_Z, raw)
            except (LarkError, ParsingExcep) as e:
                raise MalformedParamRawValue(str(e), self)
            validated = v['poly']
            poly_value = calculate(construct_instance, Poly, validated)
            sym = self.resolved_args.get('var')
            if sym:
                poly_value = poly_value.subs(poly_value.gen, sym.value)
        self.validated = validated
        self.raw = raw

        # Check irreducible.
        if not self.explicit_cyclo:
            # Careful! `poly_value.is_irreducible` is a `@property` method,
            # so must use `getattr` here as the function:
            is_irred = calculate(getattr, poly_value, 'is_irreducible')
            if not is_irred:
                msg = r'Polynomial must be irreducible over $\mathbb{Q}$.'
                raise MalformedParamRawValue(msg, self)

        i = self.resolved_args['rootIdx'].value
        root = calculate(construct_instance, CRootOf, poly_value, i)
        ext = calculate(construct_instance, AlgebraicNumber,
                        (poly_value, root),
                        alias=self.get_field_generator_symbol())
        field = calculate(QQ.algebraic_field, ext)

        return field

    def write_chooser_widget(self):
        context = {
            'descrip': self.auto_descrip(),
            'current_value': self.raw,
        }
        return nf_chooser_widget_template.render(context)

nf_chooser_widget_template = jinja2.Template("""
<div class="chooser nf_chooser">
    <div class="heading">{{ descrip }}</div>
    <div class="error_display"></div>
    <input class="textfield polynomial_field" type="text" placeholder="Irreducible polynomial" value="{{current_value}}"/>
</div>
""")



# --------------------------------------------------------------------------- #
#   Proofscape Examplorers                                                    #
#                                                                             #
#   Copyright (c) 2018-2024 Proofscape Contributors                           #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


import pytest

from sympy import Symbol

from displaylang.exceptions import CannotCall
from pfsc_examp.parse.display import displaylang_processor


build_01 = """
return str((Matrix([[2, 3], [5, 7]]) @ Matrix([[-7, 3], [5, -2]])).flat())
"""

build_02 = """
x, y, z = symbols('x, y, z')
return latex(Poly(x + y**2 + z**3).as_expr())
"""

@pytest.mark.parametrize('code, s_exp, d_exp', [
    [build_01, '[1, 0, 0, 1]', {}],
    [build_02, 'x + y^{2} + z^{3}', {'x': Symbol('x'), 'y': Symbol('y'), 'z': Symbol('z')}],
])
def test_builds(code, s_exp, d_exp):
    s, d = displaylang_processor.process(code, {})
    interactive = False
    if interactive:
        print()
        print(s)
        print(d)
    else:
        assert s == s_exp
        assert d == d_exp


@pytest.mark.parametrize('code', [
    'Poly("foo")',
    'Symbol("x[]")',
    'symbols("x, y, z()")',
])
def test_reject_str_args(code):
    code += '\nreturn "foo"'
    with pytest.raises(CannotCall) as e:
        displaylang_processor.process(code, {})
    #print(e.value)


@pytest.mark.parametrize('code', [
    'range(3)',
    'max(3, S(4), 5.12)',
    'min(3, S(4), 5.12)',
    'range(S(3))',
    'binomial(3, 2)',
    'S(3)',
    'S(1)/2',
    'S(3.14)',
    'latex(S(3))',
    # latex() on a Matrix (which is not an Expr)
    'latex(Matrix([[1, 2], [3, 4]]))',
    'mod_inverse(2, 7)',
    'x = Symbol("x"); Poly(x**2 - 2, x, modulus=7)',
    'Symbol("theta")',
    'Symbol("theta2")',
    'symbols("x, y0, z12")',
    # Construct cyclotomic field on an Integer (instead of int)
    'QQ.cyclotomic_field(S(7))',
    'k = QQ.cyclotomic_field(7); B = k.integral_basis()',
    'k = QQ.cyclotomic_field(7); P = k.primes_above(11); a = 4*P[0].ZK.parent(3); P[0].reduce_element(a)',
    'k = QQ.cyclotomic_field(7); P = k.primes_above(11); a = k([1, 2, 3]); P[0].reduce_ANP(a)',
    'k = QQ.cyclotomic_field(7); P = k.primes_above(11); a = k.to_alg_num(k([1, 2, 3])); P[0].reduce_alg_num(a)',
    'k = QQ.cyclotomic_field(7); a = k.to_sympy(k.new([1, 2, 3]))',
    'k = QQ.cyclotomic_field(7); a = k.to_sympy(k([1, 2, 3]))',
    # EvalfMixin.evalf
    'zeta = QQ.cyclotomic_field(3).ext; zeta.n(5)',
    # Expr.expand
    'x = Symbol("x"); ((x + 3)*(x - 5)).expand()',
    # Expr.subs(Expr, Expr)
    'x, y = symbols("x, y"); ((x + 3)*(x - 5)).subs(x, y)',
    # Expr.subs(Iterable)
    'x, y = symbols("x, y"); (x*y + 7).subs([(x, 1), (y, 2)], simultaneous=True)',
    # Matrix.flat()
    'Matrix([[1, 2], [3, 4]]).flat()',
    # Matrix.transpose()
    'Matrix([[1, 2], [3, 4]]).transpose()',
    # Poly.as_expr()
    'x = Symbol("x"); Poly(x**2 - 2).as_expr()',
])
def test_allow_call(code):
    """
    Just testing that we can process these code strings without raising any
    exception.
    """
    code += '\nreturn "foo"'
    displaylang_processor.process(code, {})


@pytest.mark.parametrize('code', [
    'range("foo")',
    'S("3")',
    'S("zeta")',
    'latex(S(3), order="foo()")',
    'Matrix([["1", 2], [3, 4]])',
    'mod_inverse(2.1, 7)',
    'Poly("x**2 - 2")',
    'x = Symbol("x"); Poly(x**2 - 2, 5, modulus=7)',
    'x = Symbol("x"); Poly(x**2 - 2, x, modulus=7.3)',
    'Symbol("theta[x]")',
    'symbols("x, y0, z+")',
    'k = QQ.cyclotomic_field(7.5); B = k.integral_basis(fmt="foo.bar")',
    'k = QQ.cyclotomic_field(7); B = k.integral_basis(fmt="foo.bar")',
    'k = QQ.cyclotomic_field(7); P = k.primes_above("foo")',
    'k = QQ.cyclotomic_field(7); P = k.primes_above(11); P[0].reduce_element("foo")',
    'k = QQ.cyclotomic_field(7); P = k.primes_above(11); P[0].reduce_ANP("foo")',
    'k = QQ.cyclotomic_field(7); P = k.primes_above(11); P[0].reduce_alg_num("foo")',
    'k = QQ.cyclotomic_field(7); a = k.to_sympy(k.new("foo"))',
    'k = QQ.cyclotomic_field(7); a = k.to_sympy(k("foo"))',
    'k = QQ.cyclotomic_field(7); a = k.to_sympy("foo")',
    'Matrix([[1, 2], [3, 4]]).flat("foo")',
    'x = Symbol("x"); Poly(x**2 - 2).as_expr("foo")',
])
def test_disallow_call(code):
    with pytest.raises(CannotCall):
        code += '\nreturn "foo"'
        displaylang_processor.process(code, {})
